@@ -28,33 +28,28 @@
 #include <getopt.h>
 #include <errno.h>
 #include <fcntl.h>
-#include <libintl.h>
-#include <locale.h>
-#include <dirent.h>
-#define _(s) gettext(s)
 
 /* enable the following line to get voluminous debug info */
 /* #define DEBUG */
 
 #include <unistd.h>
 #include <limits.h>
-#include <sys/sysctl.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/apparmor.h>
 
 #include "lib.h"
+#include "features.h"
 #include "parser.h"
 #include "parser_version.h"
 #include "parser_include.h"
+#include "common_optarg.h"
+#include "policy_cache.h"
 #include "libapparmor_re/apparmor_re.h"
 
-#define MODULE_NAME "apparmor"
 #define OLD_MODULE_NAME "subdomain"
 #define PROC_MODULES "/proc/modules"
-#define DEFAULT_APPARMORFS "/sys/kernel/security/" MODULE_NAME
-#define MATCH_STRING "/sys/kernel/security/" MODULE_NAME "/matching"
-#define FLAGS_FILE "/sys/kernel/security/" MODULE_NAME "/features"
+#define MATCH_FILE "/sys/kernel/security/" MODULE_NAME "/matching"
 #define MOUNTED_FS "/proc/mounts"
 #define AADFA "pattern=aadfa"
 
@@ -74,21 +69,22 @@ int skip_read_cache = 0;
 int write_cache = 0;
 int cond_clear_cache = 1;		/* only applies if write is set */
 int force_clear_cache = 0;		/* force clearing regargless of state */
-int create_cache_dir = 0;		/* create the cache dir if missing? */
+int create_cache_dir = 0;		/* DEPRECATED in favor of write_cache */
 int preprocess_only = 0;
 int skip_mode_force = 0;
-struct timespec mru_tstamp;
+int abort_on_error = 0;			/* stop processing profiles if error */
+int skip_bad_cache_rebuild = 0;
+int mru_skip_cache = 1;
+int debug_cache = 0;
+struct timespec cache_tstamp, mru_policy_tstamp;
 
-#define FLAGS_STRING_SIZE 8192
-char *match_string = NULL;
-char *flags_string = NULL;
-char *cacheloc = NULL;
+static char *apparmorfs = NULL;
+static char *cacheloc = NULL;
 
-/* per-profile settings */
-int force_complain = 0;
+static aa_features *features = NULL;
 
 /* Make sure to update BOTH the short and long_options */
-static const char *short_options = "adf:h::rRVvI:b:BCD:NSm:qQn:XKTWkL:O:po:";
+static const char *short_options = "adf:h::rRVvI:b:BCD:NSm:M:qQn:XKTWkL:O:po:";
 struct option long_options[] = {
 	{"add", 		0, 0, 'a'},
 	{"binary",		0, 0, 'B'},
@@ -106,6 +102,7 @@ struct option long_options[] = {
 	{"stdout",		0, 0, 'S'},
 	{"ofile",		1, 0, 'o'},
 	{"match-string",	1, 0, 'm'},
+	{"features-file",	1, 0, 'M'},
 	{"quiet",		0, 0, 'q'},
 	{"skip-kernel-load",	0, 0, 'Q'},
 	{"verbose",		0, 0, 'v'},
@@ -125,12 +122,16 @@ struct option long_options[] = {
 	{"optimize",		1, 0, 'O'},
 	{"Optimize",		1, 0, 'O'},
 	{"preprocess",		0, 0, 'p'},
+	{"abort-on-error",	0, 0, 132},	/* no short option */
+	{"skip-bad-cache-rebuild",	0, 0, 133},	/* no short option */
+	{"warn",		1, 0, 134},	/* no short option */
+	{"debug-cache",		0, 0, 135},	/* no short option */
 	{NULL, 0, 0, 0},
 };
 
 static int debug = 0;
 
-static void display_version(void)
+void display_version(void)
 {
 	printf("%s version " PARSER_VERSION "\n%s\n", parser_title,
 	       parser_copyright);
@@ -153,7 +154,8 @@ static void display_usage(const char *command)
 	       "-b n, --base n		Set base dir and cwd\n"
 	       "-I n, --Include n	Add n to the search path\n"
 	       "-f n, --subdomainfs n	Set location of apparmor filesystem\n"
-	       "-m n, --match-string n  Use only match features n\n"
+	       "-m n, --match-string n  Use only features n\n"
+	       "-M n, --features-file n Use only features in file n\n"
 	       "-n n, --namespace n	Set Namespace for the profile\n"
 	       "-X, --readimpliesX	Map profile read permissions to mr\n"
 	       "-k, --show-cache	Report cache hit/miss details\n"
@@ -162,7 +164,7 @@ static void display_usage(const char *command)
 	       "-W, --write-cache	Save cached profile (force with -T)\n"
 	       "    --skip-bad-cache	Don't clear cache if out of sync\n"
 	       "    --purge-cache	Clear cache regardless of its state\n"
-	       "    --create-cache-dir	Create the cache dir if missing\n"
+	       "    --debug-cache       Debug cache file checks\n"
 	       "-L, --cache-loc n	Set the location of the profile cache\n"
 	       "-q, --quiet		Don't emit warnings\n"
 	       "-v, --verbose		Show profile names as they load\n"
@@ -173,163 +175,28 @@ static void display_usage(const char *command)
 	       "-D [n], --dump		Dump internal info for debugging\n"
 	       "-O [n], --Optimize	Control dfa optimizations\n"
 	       "-h [cmd], --help[=cmd]  Display this text or info about cmd\n"
+	       "--abort-on-error	Abort processing of profiles on first error\n"
+	       "--skip-bad-cache-rebuild Do not try rebuilding the cache if it is rejected by the kernel\n"
+	       "--warn n		Enable warnings (see --help=warn)\n"
 	       ,command);
 }
 
-/*
- * flag: 1 - allow no- inversion
- * flag: 2 - flags specified should be masked off
- */
-typedef struct {
-	int control;
-	const char *option;
-	const char *desc;
-	dfaflags_t flags;
-} optflag_table_t;
-
-optflag_table_t dumpflag_table[] = {
-	{ 1, "rule-exprs", "Dump rule to expr tree conversions",
-	  DFA_DUMP_RULE_EXPR },
-	{ 1, "expr-stats", "Dump stats on expr tree", DFA_DUMP_TREE_STATS },
-	{ 1, "expr-tree", "Dump expression tree", DFA_DUMP_TREE },
-	{ 1, "expr-simplified", "Dump simplified expression tree",
-	  DFA_DUMP_SIMPLE_TREE },
-	{ 1, "stats", "Dump all compile stats",
-	  DFA_DUMP_TREE_STATS | DFA_DUMP_STATS | DFA_DUMP_TRANS_STATS |
-	  DFA_DUMP_EQUIV_STATS | DFA_DUMP_DIFF_STATS },
-	{ 1, "progress", "Dump progress for all compile phases",
-	  DFA_DUMP_PROGRESS | DFA_DUMP_STATS | DFA_DUMP_TRANS_PROGRESS |
-	  DFA_DUMP_TRANS_STATS | DFA_DUMP_DIFF_PROGRESS | DFA_DUMP_DIFF_STATS },
-	{ 1, "dfa-progress", "Dump dfa creation as in progress",
-	  DFA_DUMP_PROGRESS | DFA_DUMP_STATS },
-	{ 1, "dfa-stats", "Dump dfa creation stats", DFA_DUMP_STATS },
-	{ 1, "dfa-states", "Dump dfa state diagram", DFA_DUMP_STATES },
-	{ 1, "dfa-graph", "Dump dfa dot (graphviz) graph", DFA_DUMP_GRAPH },
-	{ 1, "dfa-minimize", "Dump dfa minimization", DFA_DUMP_MINIMIZE },
-	{ 1, "dfa-unreachable", "Dump dfa unreachable states",
-	  DFA_DUMP_UNREACHABLE },
-	{ 1, "dfa-node-map", "Dump expr node set to state mapping",
-	  DFA_DUMP_NODE_TO_DFA },
-	{ 1, "dfa-uniq-perms", "Dump unique perms",
-	  DFA_DUMP_UNIQ_PERMS },
-	{ 1, "dfa-minimize-uniq-perms", "Dump unique perms post minimization",
-	  DFA_DUMP_MIN_UNIQ_PERMS },
-	{ 1, "dfa-minimize-partitions", "Dump dfa minimization partitions",
-	  DFA_DUMP_MIN_PARTS },
-	{ 1, "compress-progress", "Dump progress of compression",
-	  DFA_DUMP_TRANS_PROGRESS | DFA_DUMP_TRANS_STATS },
-	{ 1, "compress-stats", "Dump stats on compression",
-	  DFA_DUMP_TRANS_STATS },
-	{ 1, "compressed-dfa", "Dump compressed dfa", DFA_DUMP_TRANS_TABLE },
-	{ 1, "equiv-stats", "Dump equivance class stats",
-	  DFA_DUMP_EQUIV_STATS },
-	{ 1, "equiv", "Dump equivance class", DFA_DUMP_EQUIV },
-	{ 1, "diff-encode", "Dump differential encoding",
-	  DFA_DUMP_DIFF_ENCODE },
-	{ 1, "diff-stats", "Dump differential encoding stats",
-	  DFA_DUMP_DIFF_STATS },
-	{ 1, "diff-progress", "Dump progress of differential encoding",
-	  DFA_DUMP_DIFF_PROGRESS | DFA_DUMP_DIFF_STATS },
+optflag_table_t warnflag_table[] = {
+	{ 0, "rule-not-enforced", "warn if a rule is not enforced", WARN_RULE_NOT_ENFORCED },
+	{ 0, "rule-downgraded", "warn if a rule is downgraded to a lesser but still enforcing rule", WARN_RULE_DOWNGRADED },
 	{ 0, NULL, NULL, 0 },
 };
 
-optflag_table_t optflag_table[] = {
-	{ 2, "0", "no optimizations",
-	  DFA_CONTROL_TREE_NORMAL | DFA_CONTROL_TREE_SIMPLE |
-	  DFA_CONTROL_MINIMIZE | DFA_CONTROL_REMOVE_UNREACHABLE |
-	  DFA_CONTROL_DIFF_ENCODE
-	},
-	{ 1, "equiv", "use equivalent classes", DFA_CONTROL_EQUIV },
-	{ 1, "expr-normalize", "expression tree normalization",
-	  DFA_CONTROL_TREE_NORMAL },
-	{ 1, "expr-simplify", "expression tree simplification",
-	  DFA_CONTROL_TREE_SIMPLE },
-	{ 0, "expr-left-simplify", "left simplification first",
-	  DFA_CONTROL_TREE_LEFT },
-	{ 2, "expr-right-simplify", "right simplification first",
-	  DFA_CONTROL_TREE_LEFT },
-	{ 1, "minimize", "dfa state minimization", DFA_CONTROL_MINIMIZE },
-	{ 1, "filter-deny", "filter out deny information from final dfa",
-	  DFA_CONTROL_FILTER_DENY },
-	{ 1, "remove-unreachable", "dfa unreachable state removal",
-	  DFA_CONTROL_REMOVE_UNREACHABLE },
-	{ 0, "compress-small",
-	  "do slower dfa transition table compression",
-	  DFA_CONTROL_TRANS_HIGH },
-	{ 2, "compress-fast", "do faster dfa transition table compression",
-	  DFA_CONTROL_TRANS_HIGH },
-	{ 1, "diff-encode", "Differentially encode transitions",
-	  DFA_CONTROL_DIFF_ENCODE },
-	{ 0, NULL, NULL, 0 },
-};
-
-static void print_flag_table(optflag_table_t *table)
-{
-	int i;
-	unsigned int longest = 0;
-	for (i = 0; table[i].option; i++) {
-		if (strlen(table[i].option) > longest)
-			longest = strlen(table[i].option);
-	}
-
-	for (i = 0; table[i].option; i++) {
-		printf("%5s%-*s \t%s\n", (table[i].control & 1) ? "[no-]" : "",
-		       longest, table[i].option, table[i].desc);
-	}
-}
-
-static int handle_flag_table(optflag_table_t *table, const char *optarg,
-			     dfaflags_t *flags)
-{
-	const char *arg = optarg;
-	int i, invert = 0;
-
-	if (strncmp(optarg, "no-", 3) == 0) {
-		arg = optarg + 3;
-		invert = 1;
-	}
-
-	for (i = 0; table[i].option; i++) {
-		if (strcmp(table[i].option, arg) == 0) {
-			/* check if leading no- was specified but is not
-			 * supported by the option */
-			if (invert && !(table[i].control & 1))
-				return 0;
-			if (table[i].control & 2)
-				invert |= 1;
-			if (invert)
-				*flags &= ~table[i].flags;
-			else
-				*flags |= table[i].flags;
-			return 1;
-		}
-	}
-	return 0;
-}
-
-static void display_dump(const char *command)
+void display_warn(const char *command)
 {
 	display_version();
-	printf("\n%s: --dump [Option]\n\n"
-	       "Options:\n"
-	       "--------\n"
-	       "     none specified \tDump variables\n"
-	       "     variables      \tDump variables\n"
-	       ,command);
-	print_flag_table(dumpflag_table);
-}
-
-static void display_optimize(const char *command)
-{
-	display_version();
-	printf("\n%s: -O [Option]\n\n"
+	printf("\n%s: --warn [Option]\n\n"
 	       "Options:\n"
 	       "--------\n"
 	       ,command);
-	print_flag_table(optflag_table);
+	print_flag_table(warnflag_table);
 }
 
-
 /* Treat conf file like options passed on command line
  */
 static int getopt_long_file(FILE *f, const struct option *longopts,
@@ -433,6 +300,8 @@ static int process_arg(int c, char *optarg)
 			   strcmp(optarg, "optimize") == 0 ||
 			   strcmp(optarg, "O") == 0) {
 			display_optimize(progname);
+		} else if (strcmp(optarg, "warn") == 0) {
+			display_warn(progname);
 		} else {
 			PERROR("%s: Invalid --help option %s\n",
 			       progname, optarg);
@@ -492,7 +361,7 @@ static int process_arg(int c, char *optarg)
 		}
 		break;
 	case 'f':
-		subdomainbase = strndup(optarg, PATH_MAX);
+		apparmorfs = strndup(optarg, PATH_MAX);
 		break;
 	case 'D':
 		skip_read_cache = 1;
@@ -520,11 +389,25 @@ static int process_arg(int c, char *optarg)
 		}
 		break;
 	case 'm':
-		match_string = strdup(optarg);
+		if (aa_features_new_from_string(&features,
+						optarg, strlen(optarg))) {
+			fprintf(stderr,
+				"Failed to parse features string: %m\n");
+			exit(1);
+		}
+		break;
+	case 'M':
+		if (aa_features_new(&features, AT_FDCWD, optarg)) {
+			fprintf(stderr,
+				"Failed to load features from '%s': %m\n",
+				optarg);
+			exit(1);
+		}
 		break;
 	case 'q':
 		conf_verbose = 0;
 		conf_quiet = 1;
+		warnflags = 0;
 		break;
 	case 'v':
 		conf_verbose = 1;
@@ -557,6 +440,12 @@ static int process_arg(int c, char *optarg)
 	case 131:
 		create_cache_dir = 1;
 		break;
+	case 132:
+		abort_on_error = 1;
+		break;
+	case 133:
+		skip_bad_cache_rebuild = 1;
+		break;
 	case 'L':
 		cacheloc = strdup(optarg);
 		break;
@@ -570,6 +459,17 @@ static int process_arg(int c, char *optarg)
 		preprocess_only = 1;
 		skip_mode_force = 1;
 		break;
+	case 134:
+		if (!handle_flag_table(warnflag_table, optarg,
+				       &warnflags)) {
+			PERROR("%s: Invalid --warn option %s\n",
+			       progname, optarg);
+			exit(1);
+		}
+		break;
+	case 135:
+		debug_cache = 1;
+		break;
 	default:
 		display_usage(progname);
 		exit(1);
@@ -604,7 +504,7 @@ static int process_args(int argc, char *argv[])
 static int process_config_file(const char *name)
 {
 	char *optarg;
-	FILE *f;
+	autofclose FILE *f = NULL;
 	int c, o;
 
 	f = fopen(name, "r");
@@ -613,27 +513,9 @@ static int process_config_file(const char *name)
 
 	while ((c = getopt_long_file(f, long_options, &optarg, &o)) != -1)
 		process_arg(c, optarg);
-	fclose(f);
 	return 1;
 }
 
-
-int find_subdomainfs_mountpoint(void)
-{
-	if (aa_find_mountpoint(&subdomainbase) == -1) {
-		struct stat buf;
-		if (stat(DEFAULT_APPARMORFS, &buf) == -1) {
-		PERROR(_("Warning: unable to find a suitable fs in %s, is it "
-			 "mounted?\nUse --subdomainfs to override.\n"),
-		       MOUNTED_FS);
-		} else {
-			subdomainbase = strdup(DEFAULT_APPARMORFS);
-		}
-	}
-
-	return (subdomainbase == NULL);
-}
-
 int have_enough_privilege(void)
 {
 	uid_t uid, euid;
@@ -657,250 +539,118 @@ int have_enough_privilege(void)
 	return 0;
 }
 
-char *snprintf_buffer(char *buf, char *pos, ssize_t size, const char *fmt, ...)
+static void set_features_by_match_file(void)
 {
-	va_list args;
-	int i, remaining = size - (pos - buf);
-
-	va_start(args, fmt);
-	i = vsnprintf(pos, remaining, fmt, args);
-	va_end(args);
-
-	if (i >= size) {
-		PERROR(_("Feature buffer full."));
-		exit(1);
-	}
-
-	return pos + i;
-}
-
-struct features_struct {
-	char **buffer;
-	int size;
-	char *pos;
-};
-
-static int features_dir_cb(DIR *dir, const char *name, struct stat *st,
-			   void *data)
-{
-	struct features_struct *fst = (struct features_struct *) data;
-
-	/* skip dot files and files with no name */
-	if (*name == '.' || !strlen(name))
-		return 0;
-
-	fst->pos = snprintf_buffer(*fst->buffer, fst->pos, fst->size, "%s {", name);
-
-	if (S_ISREG(st->st_mode)) {
-		int len, file;
-		int remaining = fst->size - (fst->pos - *fst->buffer);
-		if (!(file = openat(dirfd(dir), name, O_RDONLY))) {
-			PDEBUG("Could not open '%s'", name);
-			return -1;
-		}
-		PDEBUG("Opened features \"%s\"\n", name);
-		if (st->st_size > remaining) {
-			PDEBUG("Feature buffer full.");
-			return -1;
-		}
-
-		do {
-			len = read(file, fst->pos, remaining);
-			if (len > 0) {
-				remaining -= len;
-				fst->pos += len;
-				*fst->pos = 0;
-			}
-		} while (len > 0);
-		if (len < 0) {
-			PDEBUG("Error reading feature file '%s'\n", name);
-			return -1;
-		}
-		close(file);
-	} else if (S_ISDIR(st->st_mode)) {
-		if (dirat_for_each(dir, name, fst, features_dir_cb))
-			return -1;
-	}
-
-	fst->pos = snprintf_buffer(*fst->buffer, fst->pos, fst->size, "}\n");
-
-	return 0;
-}
-
-static char *handle_features_dir(const char *filename, char **buffer, int size,
-				 char *pos)
-{
-	struct features_struct fst = { buffer, size, pos };
-
-	if (dirat_for_each(NULL, filename, &fst, features_dir_cb)) {
-		PDEBUG("Failed evaluating .features\n");
-		exit(1);
-	}
-
-	return fst.pos;
-}
-
-/* match_string == NULL --> no match_string available
-   match_string != NULL --> either a matching string specified on the
-   command line, or the kernel supplied a match string */
-static void get_match_string(void) {
-
-	FILE *ms = NULL;
-	struct stat stat_file;
-
-	/* has process_args() already assigned a match string? */
-	if (match_string)
-		goto out;
-
-	if (stat(FLAGS_FILE, &stat_file) == -1)
-		goto out;
-
-	if (S_ISDIR(stat_file.st_mode)) {
-		/* if we have a features directory default to */
-		perms_create = 1;
-
-		flags_string = (char *) malloc(FLAGS_STRING_SIZE);
-		handle_features_dir(FLAGS_FILE, &flags_string, FLAGS_STRING_SIZE, flags_string);
-		if (strstr(flags_string, "network"))
-			kernel_supports_network = 1;
-		else
-			kernel_supports_network = 0;
-		if (strstr(flags_string, "mount"))
-			kernel_supports_mount = 1;
-		if (strstr(flags_string, "dbus"))
-			kernel_supports_dbus = 1;
-		return;
-	}
-
-	ms = fopen(MATCH_STRING, "r");
-	if (!ms)
-		goto out;
-
-	match_string = (char *) malloc(1000);
-	if (!match_string) {
-		goto out;
-	}
-
-	if (!fgets(match_string, 1000, ms)) {
-		free(match_string);
-		match_string = NULL;
-	}
-
-out:
-	if (match_string) {
+	autofclose FILE *ms = fopen(MATCH_FILE, "r");
+	if (ms) {
+		autofree char *match_string = (char *) malloc(1000);
+		if (!match_string)
+			goto no_match;
+		if (!fgets(match_string, 1000, ms))
+			goto no_match;
 		if (strstr(match_string, " perms=c"))
 			perms_create = 1;
-	} else {
-		perms_create = 1;
-		kernel_supports_network = 0;
+		kernel_supports_network = 1;
+		return;
 	}
-
-	if (ms)
-		fclose(ms);
-	return;
+no_match:
+	perms_create = 1;
 }
 
-static void get_flags_string(char **flags, const char *flags_file) {
-	char *pos;
-	FILE *f = NULL;
-	size_t size;
-
-	/* abort if missing or already set */
-	if (!flags || *flags)
-		return;
-
-	f = fopen(flags_file, "r");
-	if (!f)
-		return;
-
-	*flags = (char *) malloc(FLAGS_STRING_SIZE);
-	if (!*flags)
-		goto fail;
-
-	size = fread(*flags, 1, FLAGS_STRING_SIZE - 1, f);
-	if (!size || ferror(f))
-		goto fail;
-	(*flags)[size] = 0;
-
-	fclose(f);
-	return;
-
-fail:
-	free(*flags);
-	*flags = NULL;
-	if (f)
-		fclose(f);
-	return;
-}
-
-int process_binary(int option, const char *profilename)
+static void set_supported_features(void)
 {
-	char *buffer = NULL;
-	int retval = 0, size = 0, asize = 0, rsize;
-	int chunksize = 1 << 14;
-	int fd;
-
-	if (profilename) {
-		fd = open(profilename, O_RDONLY);
-		if (fd == -1) {
-			PERROR(_("Error: Could not read profile %s: %s.\n"),
-			       profilename, strerror(errno));
-			exit(errno);
-		}
-	} else {
-		fd = dup(0);
+	/* has process_args() already assigned a match string? */
+	if (!features && aa_features_new_from_kernel(&features) == -1) {
+		set_features_by_match_file();
+		return;
 	}
 
-	do {
-		if (asize - size == 0) {
-		  buffer = (char *) realloc(buffer, chunksize);
-			asize = chunksize;
-			chunksize <<= 1;
-			if (!buffer) {
-				PERROR(_("Memory allocation error."));
-				exit(errno);
+	perms_create = 1;
+	kernel_supports_policydb = aa_features_supports(features, "file");
+	kernel_supports_network = aa_features_supports(features, "network");
+	kernel_supports_unix = aa_features_supports(features,
+						    "network/af_unix");
+	kernel_supports_mount = aa_features_supports(features, "mount");
+	kernel_supports_dbus = aa_features_supports(features, "dbus");
+	kernel_supports_signal = aa_features_supports(features, "signal");
+	kernel_supports_ptrace = aa_features_supports(features, "ptrace");
+	kernel_supports_setload = aa_features_supports(features,
+						       "policy/set_load");
+	kernel_supports_diff_encode = aa_features_supports(features,
+							   "policy/diff_encode");
+
+	if (aa_features_supports(features, "policy/versions/v7"))
+		kernel_abi_version = 7;
+	else if (aa_features_supports(features, "policy/versions/v6"))
+		kernel_abi_version = 6;
+
+	if (!kernel_supports_diff_encode)
+		/* clear diff_encode because it is not supported */
+		dfaflags &= ~DFA_CONTROL_DIFF_ENCODE;
+}
+
+int process_binary(int option, aa_kernel_interface *kernel_interface,
+		   const char *profilename)
+{
+	const char *printed_name;
+	int retval;
+
+	printed_name = profilename ? profilename : "stdin";
+
+	if (kernel_load) {
+		if (option == OPTION_ADD) {
+			retval = profilename ?
+				 aa_kernel_interface_load_policy_from_file(kernel_interface, AT_FDCWD, profilename) :
+				 aa_kernel_interface_load_policy_from_fd(kernel_interface, 0);
+			if (retval == -1) {
+				retval = errno;
+				PERROR(_("Error: Could not load profile %s: %s\n"),
+				       printed_name, strerror(retval));
+				return retval;
 			}
+		} else if (option == OPTION_REPLACE) {
+			retval = profilename ?
+				 aa_kernel_interface_replace_policy_from_file(kernel_interface, AT_FDCWD, profilename) :
+				 aa_kernel_interface_replace_policy_from_fd(kernel_interface, 0);
+			if (retval == -1) {
+				retval = errno;
+				PERROR(_("Error: Could not replace profile %s: %s\n"),
+				       printed_name, strerror(retval));
+				return retval;
+			}
+		} else {
+			PERROR(_("Error: Invalid load option specified: %d\n"),
+			       option);
+			return EINVAL;
 		}
-
-		rsize = read(fd, buffer + size, asize - size);
-		if (rsize)
-			size += rsize;
-	} while (rsize > 0);
-
-	close(fd);
-
-	if (rsize == 0)
-		retval = sd_load_buffer(option, buffer, size);
-	else
-		retval = rsize;
-
-	free(buffer);
+	}
 
 	if (conf_verbose) {
 		switch (option) {
 		case OPTION_ADD:
 			printf(_("Cached load succeeded for \"%s\".\n"),
-			       profilename ? profilename : "stdin");
+			       printed_name);
 			break;
 		case OPTION_REPLACE:
 			printf(_("Cached reload succeeded for \"%s\".\n"),
-			       profilename ? profilename : "stdin");
+			       printed_name);
 			break;
 		default:
 			break;
 		}
 	}
 
-	return retval;
+	return 0;
 }
 
 void reset_parser(const char *filename)
 {
-	memset(&mru_tstamp, 0, sizeof(mru_tstamp));
+	memset(&mru_policy_tstamp, 0, sizeof(mru_policy_tstamp));
+	memset(&cache_tstamp, 0, sizeof(cache_tstamp));
+	mru_skip_cache = 1;
 	free_aliases();
 	free_symtabs();
 	free_policies();
-	reset_regex();
 	reset_include_stack(filename);
 }
 
@@ -909,7 +659,7 @@ int test_for_dir_mode(const char *basename, const char *linkdir)
 	int rc = 0;
 
 	if (!skip_mode_force) {
-		char *target = NULL;
+		autofree char *target = NULL;
 		if (asprintf(&target, "%s/%s/%s", basedir, linkdir, basename) < 0) {
 			perror("asprintf");
 			exit(1);
@@ -917,36 +667,19 @@ int test_for_dir_mode(const char *basename, const char *linkdir)
 
 		if (access(target, R_OK) == 0)
 			rc = 1;
-
-		free(target);
 	}
 
 	return rc;
 }
 
-
-/* returns true if time is more recent than mru_tstamp */
-#define mru_t_cmp(a) \
-(((a).tv_sec == (mru_tstamp).tv_sec) ? \
-  (a).tv_nsec > (mru_tstamp).tv_nsec : (a).tv_sec > (mru_tstamp).tv_sec)
-
-void update_mru_tstamp(FILE *file)
+int process_profile(int option, aa_kernel_interface *kernel_interface,
+		    const char *profilename, const char *cachedir)
 {
-	struct stat stat_file;
-	if (fstat(fileno(file), &stat_file))
-		return;
-	if (mru_t_cmp(stat_file.st_ctim))
-		mru_tstamp = stat_file.st_ctim;
-}
-
-int process_profile(int option, const char *profilename)
-{
-	struct stat stat_bin;
 	int retval = 0;
-	char * cachename = NULL;
-	char * cachetemp = NULL;
+	autofree const char *cachename = NULL;
+	autofree const char *cachetmpname = NULL;
+	autoclose int cachetmp = -1;
 	const char *basename = NULL;
-	FILE *cmd;
 
 	/* per-profile states */
 	force_complain = opt_force_complain;
@@ -955,13 +688,14 @@ int process_profile(int option, const char *profilename)
 		if ( !(yyin = fopen(profilename, "r")) ) {
 			PERROR(_("Error: Could not read profile %s: %s.\n"),
 			       profilename, strerror(errno));
-			exit(errno);
+			return errno;
 		}
-	}
-	else {
+	} else {
 		pwarn("%s: cannot use or update cache, disable, or force-complain via stdin\n", progname);
 	}
 
+	reset_parser(profilename);
+
 	if (profilename && option != OPTION_REMOVE) {
 		/* make decisions about disabled or complain-mode profiles */
 		basename = strrchr(profilename, '/');
@@ -981,23 +715,17 @@ int process_profile(int option, const char *profilename)
  			force_complain = 1;
  		}
 
-		/* TODO: add primary cache check.
-		 * If .file for cached binary exists get the list of profile
-		 * names and check their time stamps.
-		 */
-		/* TODO: primary cache miss/hit messages */
+		/* setup cachename and tstamp */
+		if (!force_complain && !skip_cache) {
+			cachename = cache_filename(cachedir, basename);
+			valid_read_cache(cachename);
+		}
+
 	}
 
-	reset_parser(profilename);
 	if (yyin) {
 		yyrestart(yyin);
-		update_mru_tstamp(yyin);
-	}
-
-	cmd = fopen("/proc/self/exe", "r");
-	if (cmd) {
-		update_mru_tstamp(cmd);
-		fclose(cmd);
+		update_mru_tstamp(yyin, profilename ? profilename : "stdin");
 	}
 
 	retval = yyparse();
@@ -1009,42 +737,21 @@ int process_profile(int option, const char *profilename)
 	 * TODO: Add support for caching profiles in an alternate namespace
 	 * TODO: Add support for embedded namespace defines if they aren't
 	 *       removed from the language.
+	 * TODO: test profile->ns NOT profile_ns (must be after parse)
 	 */
 	if (profile_ns)
 		skip_cache = 1;
 
-	/* Do secondary test to see if cached binary profile is good,
-	 * instead of checking against a presupplied list of files
-	 * use the timestamps from the files that were parsed.
-	 * Parsing the profile is slower that doing primary cache check
-	 * its still faster than doing full compilation
-	 */
-	if ((profilename && option != OPTION_REMOVE) && !force_complain &&
-	    !skip_cache) {
-		if (asprintf(&cachename, "%s/%s", cacheloc, basename)<0) {
-			PERROR(_("Memory allocation error."));
-			exit(1);
-		}
+	if (cachename) {
 		/* Load a binary cache if it exists and is newest */
-		if (!skip_read_cache &&
-		    stat(cachename, &stat_bin) == 0 &&
-		    stat_bin.st_size > 0 && (mru_t_cmp(stat_bin.st_mtim))) {
-			if (show_cache)
-				PERROR("Cache hit: %s\n", cachename);
-			retval = process_binary(option, cachename);
-			goto out;
-		}
-		if (write_cache) {
-			/* Otherwise, set up to save a cached copy */
-			if (asprintf(&cachetemp, "%s-XXXXXX", cachename)<0) {
-				perror("asprintf");
-				exit(1);
-			}
-			if ( (cache_fd = mkstemp(cachetemp)) < 0) {
-				perror("mkstemp");
-				exit(1);
-			}
+		if (cache_hit(cachename)) {
+			retval = process_binary(option, kernel_interface,
+						cachename);
+			if (!retval || skip_bad_cache_rebuild)
+				return retval;
 		}
+
+		cachetmp = setup_cache_tmp(&cachetmpname, cachename);
 	}
 
 	if (show_cache)
@@ -1080,192 +787,81 @@ int process_profile(int option, const char *profilename)
 		goto out;
 	}
 
-	retval = load_policy(option);
-
-out:
-	if (cachetemp) {
-		/* Only install the generate cache file if it parsed correctly
-                   and did not have write/close errors */
-		int useable_cache = (cache_fd != -1 && retval == 0);
-		if (cache_fd != -1) {
-			if (close(cache_fd))
-				useable_cache = 0;
-			cache_fd = -1;
+	/* cache file generated by load_policy */
+	retval = load_policy(option, kernel_interface, cachetmp);
+	if (retval == 0 && write_cache) {
+		if (cachetmp == -1) {
+			unlink(cachetmpname);
+			PERROR("Warning failed to create cache: %s\n",
+			       basename);
+		} else {
+			install_cache(cachetmpname, cachename);
 		}
-
-		if (useable_cache) {
-			if (rename(cachetemp, cachename) < 0) {
-				pwarn("Warning failed to write cache: %s\n", cachename);
-				unlink(cachetemp);
-			}
-			else if (show_cache)
-				PERROR("Wrote cache: %s\n", cachename);
-		}
-		else {
-			unlink(cachetemp);
-			PERROR("Warning failed to create cache: %s\n", basename);
-		}
-		free(cachetemp);
 	}
-	if (cachename)
-		free(cachename);
+out:
+
 	return retval;
 }
 
-/* data - name of parent dir */
-static int profile_dir_cb(__unused DIR *dir, const char *name, struct stat *st,
+struct dir_cb_data {
+	aa_kernel_interface *kernel_interface;
+	const char *dirname;	/* name of the parent dir */
+	const char *cachedir;	/* path to the cache sub directory */
+};
+
+/* data - pointer to a dir_cb_data */
+static int profile_dir_cb(int dirfd unused, const char *name, struct stat *st,
 			  void *data)
 {
 	int rc = 0;
 
 	if (!S_ISDIR(st->st_mode) && !is_blacklisted(name, NULL)) {
-		const char *dirname = (const char *)data;
-		char *path;
-		if (asprintf(&path, "%s/%s", dirname, name) < 0)
+		struct dir_cb_data *cb_data = (struct dir_cb_data *)data;
+		autofree char *path = NULL;
+		if (asprintf(&path, "%s/%s", cb_data->dirname, name) < 0)
 			PERROR(_("Out of memory"));
-		rc = process_profile(option, path);
-		free(path);
+		rc = process_profile(option, cb_data->kernel_interface, path,
+				     cb_data->cachedir);
 	}
 	return rc;
 }
 
-/* data - name of parent dir */
-static int binary_dir_cb(__unused DIR *dir, const char *name, struct stat *st,
+/* data - pointer to a dir_cb_data */
+static int binary_dir_cb(int dirfd unused, const char *name, struct stat *st,
 			 void *data)
 {
 	int rc = 0;
 
 	if (!S_ISDIR(st->st_mode) && !is_blacklisted(name, NULL)) {
-		const char *dirname = (const char *)data;
-		char *path;
-		if (asprintf(&path, "%s/%s", dirname, name) < 0)
+		struct dir_cb_data *cb_data = (struct dir_cb_data *)data;
+		autofree char *path = NULL;
+		if (asprintf(&path, "%s/%s", cb_data->dirname, name) < 0)
 			PERROR(_("Out of memory"));
-		rc = process_binary(option, path);
-		free(path);
+		rc = process_binary(option, cb_data->kernel_interface, path);
 	}
 	return rc;
 }
 
-static int clear_cache_cb(DIR *dir, const char *path, struct stat *st,
-			  __unused void *data)
-{
-	/* remove regular files */
-	if (S_ISREG(st->st_mode))
-		return unlinkat(dirfd(dir), path, 0);
-
-	/* do nothing with other file types */
-	return 0;
-}
-
-static int clear_cache_files(const char *path)
-{
-	return dirat_for_each(NULL, path, NULL, clear_cache_cb);
-}
-
-static int create_cache(const char *cachedir, const char *path,
-			const char *features)
-{
-	struct stat stat_file;
-	FILE * f = NULL;
-
-	if (clear_cache_files(cacheloc) != 0)
-		goto error;
-
-create_file:
-	f = fopen(path, "w");
-	if (f) {
-		if (fwrite(features, strlen(features), 1, f) != 1 )
-			goto error;
-
-		fclose(f);
-
-
-		return 0;
-	}
-
-error:
-	/* does the dir exist? */
-	if (stat(cachedir, &stat_file) == -1 && create_cache_dir) {
-		if (mkdir(cachedir, 0700) == 0)
-			goto create_file;
-		if (show_cache)
-			PERROR(_("Can't create cache directory: %s\n"), cachedir);
-	} else if (!S_ISDIR(stat_file.st_mode)) {
-		if (show_cache)
-			PERROR(_("File in cache directory location: %s\n"), cachedir);
-	} else {
-		if (show_cache)
-			PERROR(_("Can't update cache directory: %s\n"), cachedir);
-	}
-
-	if (show_cache)
-		PERROR("Cache write disabled: cannot create %s\n", path);
-	write_cache = 0;
-
-	return -1;
-}
-
 static void setup_flags(void)
 {
-	char *cache_features_path = NULL;
-	char *cache_flags = NULL;
-
 	/* Get the match string to determine type of regex support needed */
-	get_match_string();
-	/* Get kernel features string */
-	get_flags_string(&flags_string, FLAGS_FILE);
+	set_supported_features();
+
 	/* Gracefully handle AppArmor kernel without compatibility patch */
-	if (!flags_string) {
-		PERROR("Cache read/write disabled: %s interface file missing. "
-			"(Kernel needs AppArmor 2.4 compatibility patch.)\n",
-			FLAGS_FILE);
+	if (!features) {
+		PERROR("Cache read/write disabled: interface file missing. "
+			"(Kernel needs AppArmor 2.4 compatibility patch.)\n");
 		write_cache = 0;
 		skip_read_cache = 1;
 		return;
-	} else if (strstr(flags_string, "network"))
-		kernel_supports_network = 1;
-	else
-		kernel_supports_network = 0;
-
-
-
-	/*
-         * Deal with cache directory versioning:
-         *  - If cache/.features is missing, create it if --write-cache.
-         *  - If cache/.features exists, and does not match flags_string,
-         *    force cache reading/writing off.
-         */
-	if (asprintf(&cache_features_path, "%s/.features", cacheloc) == -1) {
-		PERROR(_("Memory allocation error."));
-		exit(1);
 	}
-
-	get_flags_string(&cache_flags, cache_features_path);
-	if (cache_flags) {
-		if (strcmp(flags_string, cache_flags) != 0) {
-			if (write_cache && cond_clear_cache) {
-				if (create_cache(cacheloc, cache_features_path,
-						 flags_string))
-					skip_read_cache = 1;
-			} else {
-				if (show_cache)
-					PERROR("Cache read/write disabled: %s does not match %s\n", FLAGS_FILE, cache_features_path);
-				write_cache = 0;
-				skip_read_cache = 1;
-			}
-		}
-		free(cache_flags);
-		cache_flags = NULL;
-	} else if (write_cache) {
-		create_cache(cacheloc, cache_features_path, flags_string);
-	}
-
-	free(cache_features_path);
 }
 
 int main(int argc, char *argv[])
 {
-	int retval;
+	aa_kernel_interface *kernel_interface = NULL;
+	aa_policy_cache *policy_cache;
+	int retval, last_error;
 	int i;
 	int optind;
 
@@ -1287,34 +883,72 @@ int main(int argc, char *argv[])
 		return retval;
 	}
 
-	/* create the cacheloc once and use it everywhere */
-	if (!cacheloc) {
-		if (asprintf(&cacheloc, "%s/cache", basedir) == -1) {
-			PERROR(_("Memory allocation error."));
-			exit(1);
-		}
-	}
-
-	if (force_clear_cache) 
-		exit(clear_cache_files(cacheloc));
-
-	/* Check to make sure there is an interface to load policy */
-	if (!(UNPRIVILEGED_OPS) && (subdomainbase == NULL) &&
-	    (retval = find_subdomainfs_mountpoint())) {
-		return retval;
-	}
-
 	if (!binary_input) parse_default_paths();
 
 	setup_flags();
 
-	retval = 0;
-	for (i = optind; retval == 0 && i <= argc; i++) {
+	if (!(UNPRIVILEGED_OPS) &&
+	    aa_kernel_interface_new(&kernel_interface, features, apparmorfs) == -1) {
+		PERROR(_("Warning: unable to find a suitable fs in %s, is it "
+		       "mounted?\nUse --subdomainfs to override.\n"),
+		       MOUNTED_FS);
+		return 1;
+	}
+
+	if ((!skip_cache && (write_cache || !skip_read_cache)) ||
+	    force_clear_cache) {
+		uint16_t max_caches = write_cache && cond_clear_cache ? 1 : 0;
+
+		if (!cacheloc && asprintf(&cacheloc, "%s/cache", basedir) == -1) {
+			PERROR(_("Memory allocation error."));
+			return 1;
+		}
+
+		if (force_clear_cache) {
+			if (aa_policy_cache_remove(AT_FDCWD, cacheloc)) {
+				PERROR(_("Failed to clear cache files (%s): %s\n"),
+				       cacheloc, strerror(errno));
+				return 1;
+			}
+
+			return 0;
+		}
+
+		if (create_cache_dir)
+			pwarn(_("The --create-cache-dir option is deprecated. Please use --write-cache.\n"));
+
+		retval = aa_policy_cache_new(&policy_cache, features,
+					     AT_FDCWD, cacheloc, max_caches);
+		if (retval) {
+			if (errno != ENOENT && errno != EEXIST) {
+				PERROR(_("Failed setting up policy cache (%s): %s\n"),
+				       cacheloc, strerror(errno));
+				return 1;
+			}
+
+			if (show_cache) {
+				if (max_caches > 0)
+					PERROR("Cache write disabled: Cannot create cache '%s': %m\n",
+					       cacheloc);
+				else
+					PERROR("Cache read/write disabled: Policy cache is invalid\n");
+			}
+
+			write_cache = 0;
+			skip_read_cache = 1;
+		}
+	}
+
+	retval = last_error = 0;
+	for (i = optind; i <= argc; i++) {
 		struct stat stat_file;
 
 		if (i < argc && !(profilename = strdup(argv[i]))) {
 			perror("strdup");
-			return -1;
+			last_error = ENOMEM;
+			if (abort_on_error)
+				break;
+			continue;
 		}
 		/* skip stdin if we've seen other command line arguments */
 		if (i == argc && optind != argc)
@@ -1326,26 +960,38 @@ int main(int argc, char *argv[])
 		}
 
 		if (profilename && S_ISDIR(stat_file.st_mode)) {
-			int (*cb)(DIR *dir, const char *name, struct stat *st,
+			int (*cb)(int dirfd, const char *name, struct stat *st,
 				  void *data);
+			struct dir_cb_data cb_data;
+
+			cb_data.dirname = profilename;
+			cb_data.cachedir = cacheloc;
 			cb = binary_input ? binary_dir_cb : profile_dir_cb;
-			if (dirat_for_each(NULL, profilename, profilename, cb)) {
+			if ((retval = dirat_for_each(AT_FDCWD, profilename,
+						     &cb_data, cb))) {
 				PDEBUG("Failed loading profiles from %s\n",
 				       profilename);
-				exit(1);
 			}
 		} else if (binary_input) {
-			retval = process_binary(option, profilename);
+			retval = process_binary(option, kernel_interface,
+						profilename);
 		} else {
-			retval = process_profile(option, profilename);
+			retval = process_profile(option, kernel_interface,
+						 profilename, cacheloc);
 		}
 
 		if (profilename) free(profilename);
 		profilename = NULL;
+
+		if (retval) {
+			last_error = retval;
+			if (abort_on_error)
+				break;
+		}
 	}
 
 	if (ofile)
 		fclose(ofile);
 
-	return retval;
+	return last_error;
 }
