@@ -254,6 +254,11 @@ static int process_variables_in_entries(struct cod_entry *entry_list)
 		error = expand_entry_variables(&entry->name);
 		if (error)
 			return error;
+		if (entry->link_name) {
+			error = expand_entry_variables(&entry->link_name);
+			if (error)
+				return error;
+		}
 	}
 
 	return 0;
@@ -270,14 +275,51 @@ static int process_variables_in_rules(Profile &prof)
 	return 0;
 }
 
+static int process_variables_in_name(Profile &prof)
+{
+	/* this needs to be done before alias expansion, ie. altnames are
+	 * setup
+	 */
+	int error = expand_entry_variables(&prof.name);
+	if (!error && prof.attachment)
+		error = expand_entry_variables(&prof.attachment);
+
+	return error;
+}
+
+static std::string escape_re(std::string str)
+{
+	for (size_t i = 0; i < str.length(); i++) {
+		if (str[i] == '\\') {
+			/* skip \ and follow char. Skipping \ and first
+			 * char is enough for multichar escape sequence
+			 */
+			i++;
+			continue;
+		}
+		if (strchr("{}[]*?", str[i]) != NULL) {
+			str.insert(i++, "\\");
+		}
+	}
+
+	return str;
+}
 
 int process_profile_variables(Profile *prof)
 {
 	int error = 0, rc;
-	std::string *buf = prof->get_name(true);
 
-	error = new_set_var(PROFILE_NAME_VARIABLE, buf->c_str());
-	delete buf;
+	/* needs to be before PROFILE_NAME_VARIABLE so that variable will
+	 * have the correct name
+	 */
+	error = process_variables_in_name(*prof);
+
+	if (!error) {
+		/* escape profile name elements that could be interpreted
+		 * as regular expressions.
+		 */
+		error = new_set_var(PROFILE_NAME_VARIABLE, escape_re(prof->get_name(false)).c_str());
+	}
 
 	if (!error)
 		error = process_variables_in_entries(prof->entries);
