@@ -11,15 +11,14 @@
 #    GNU General Public License for more details.
 #
 # ----------------------------------------------------------------------
-import re
-
-import apparmor
+import apparmor.aa as apparmor
+from apparmor.regex import re_match_include
 
 class Prof(object):
     def __init__(self, filename):
-        self.aa = apparmor.aa.aa
-        self.filelist = apparmor.aa.filelist
-        self.include = apparmor.aa.include
+        self.aa = apparmor.aa
+        self.filelist = apparmor.filelist
+        self.include = apparmor.include
         self.filename = filename
 
 class CleanProf(object):
@@ -48,7 +47,7 @@ class CleanProf(object):
         #Process every hat in the profile individually
         file_includes = list(self.profile.filelist[self.profile.filename]['include'].keys())
         deleted = 0
-        for hat in self.profile.aa[program].keys():
+        for hat in sorted(self.profile.aa[program].keys()):
             #The combined list of includes from profile and the file
             includes = list(self.profile.aa[program][hat]['include'].keys()) + file_includes
 
@@ -61,22 +60,22 @@ class CleanProf(object):
             #Clean up superfluous rules from includes in the other profile
             for inc in includes:
                 if not self.profile.include.get(inc, {}).get(inc, False):
-                    apparmor.aa.load_include(inc)
-                deleted += apparmor.aa.delete_duplicates(self.other.aa[program][hat], inc)
+                    apparmor.load_include(inc)
+                deleted += apparmor.delete_duplicates(self.other.aa[program][hat], inc)
 
-            #Clean the duplicates of caps in other profile
-            deleted += delete_cap_duplicates(self.profile.aa[program][hat]['allow']['capability'], self.other.aa[program][hat]['allow']['capability'], self.same_file)
-            deleted += delete_cap_duplicates(self.profile.aa[program][hat]['deny']['capability'], self.other.aa[program][hat]['deny']['capability'], self.same_file)
+            #Clean duplicate rules in other profile
+            if not self.same_file:
+                deleted += self.other.aa[program][hat]['capability'].delete_duplicates(self.profile.aa[program][hat]['capability'])
+                deleted += self.other.aa[program][hat]['network'].delete_duplicates(self.profile.aa[program][hat]['network'])
+            else:
+                deleted += self.other.aa[program][hat]['capability'].delete_duplicates(None)
+                deleted += self.other.aa[program][hat]['network'].delete_duplicates(None)
 
             #Clean the duplicates of path in other profile
             deleted += delete_path_duplicates(self.profile.aa[program][hat], self.other.aa[program][hat], 'allow', self.same_file)
             deleted += delete_path_duplicates(self.profile.aa[program][hat], self.other.aa[program][hat], 'deny', self.same_file)
 
-            #Clean the duplicates of net rules in other profile
-            deleted += delete_net_duplicates(self.profile.aa[program][hat]['allow']['netdomain'], self.other.aa[program][hat]['allow']['netdomain'], self.same_file)
-            deleted += delete_net_duplicates(self.profile.aa[program][hat]['deny']['netdomain'], self.other.aa[program][hat]['deny']['netdomain'], self.same_file)
-
-            return deleted
+        return deleted
 
 def delete_path_duplicates(profile, profile_other, allow, same_profile=True):
     deleted = []
@@ -88,19 +87,19 @@ def delete_path_duplicates(profile, profile_other, allow, same_profile=True):
                 cm = profile[allow]['path'][rule]['mode']
                 am = profile[allow]['path'][rule]['audit']
                 # If modes of rule are a superset of rules implied by entry we can safely remove it
-                if apparmor.aa.mode_contains(cm, profile_other[allow]['path'][entry]['mode']) and apparmor.aa.mode_contains(am, profile_other[allow]['path'][entry]['audit']):
+                if apparmor.mode_contains(cm, profile_other[allow]['path'][entry]['mode']) and apparmor.mode_contains(am, profile_other[allow]['path'][entry]['audit']):
                     if not same_profile:
                         deleted.append(entry)
                 continue
-            if re.search('#?\s*include', rule) or re.search('#?\s*include', entry):
+            if re_match_include(rule) or re_match_include(entry):
                 continue
             # Check if the rule implies entry
-            if apparmor.aa.matchliteral(rule, entry):
+            if apparmor.matchliteral(rule, entry):
                 # Check the modes
                 cm = profile[allow]['path'][rule]['mode']
                 am = profile[allow]['path'][rule]['audit']
                 # If modes of rule are a superset of rules implied by entry we can safely remove it
-                if apparmor.aa.mode_contains(cm, profile_other[allow]['path'][entry]['mode']) and apparmor.aa.mode_contains(am, profile_other[allow]['path'][entry]['audit']):
+                if apparmor.mode_contains(cm, profile_other[allow]['path'][entry]['mode']) and apparmor.mode_contains(am, profile_other[allow]['path'][entry]['audit']):
                     deleted.append(entry)
 
     for entry in deleted:
@@ -108,44 +107,3 @@ def delete_path_duplicates(profile, profile_other, allow, same_profile=True):
 
     return len(deleted)
 
-def delete_cap_duplicates(profilecaps, profilecaps_other, same_profile=True):
-    deleted = []
-    if profilecaps and profilecaps_other and not same_profile:
-        for capname in profilecaps.keys():
-            if profilecaps_other[capname].get('set', False):
-                deleted.append(capname)
-        for capname in deleted:
-            profilecaps_other.pop(capname)
-
-    return len(deleted)
-
-def delete_net_duplicates(netrules, netrules_other, same_profile=True):
-    deleted = 0
-    hasher_obj = apparmor.aa.hasher()
-    if netrules_other and netrules:
-        netglob = False
-        # Delete matching rules
-        if netrules.get('all', False):
-            netglob = True
-        # Iterate over a copy of the rules in the other profile
-        for fam in list(netrules_other['rule'].keys()):
-            if netglob or (type(netrules['rule'][fam]) != type(hasher_obj) and netrules['rule'][fam]):
-                if not same_profile:
-                    if type(netrules_other['rule'][fam]) == type(hasher_obj):
-                        deleted += len(netrules_other['rule'][fam].keys())
-                    else:
-                        deleted += 1
-                    netrules_other['rule'].pop(fam)
-            elif type(netrules_other['rule'][fam]) != type(hasher_obj) and netrules_other['rule'][fam]:
-                if type(netrules['rule'][fam]) != type(hasher_obj) and netrules['rule'][fam]:
-                    if not same_profile:
-                        netrules_other['rule'].pop(fam)
-                        deleted += 1
-            else:
-                for sock_type in list(netrules_other['rule'][fam].keys()):
-                    if netrules['rule'].get(fam, False):
-                        if netrules['rule'][fam].get(sock_type, False):
-                            if not same_profile:
-                                netrules_other['rule'][fam].pop(sock_type)
-                                deleted += 1
-    return deleted
